package org.kantega.jexmec.felix;

/*
 * Copyright 2010 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.apache.felix.framework.Felix;
import org.apache.felix.framework.util.FelixConstants;
import org.apache.felix.main.AutoActivator;
import org.kantega.jexmec.PluginClassLoaderProvider;
import org.kantega.jexmec.PluginClassLoaderRegistry;
import org.osgi.framework.*;

import java.io.IOException;
import java.net.URL;
import java.util.*;

/**
 * Implementation of {@link org.kantega.jexmec.PluginLoader} that
 * discovers and removes class loaders from OGSI bundles as they are started and stopped.
 */
public class FelixPluginClassLoaderProvider implements PluginClassLoaderProvider {
    private Felix m_felix;

    private List<String> systemPackages;

    public FelixPluginClassLoaderProvider(List<String> systemPackages) {
        this.systemPackages = systemPackages;
    }

    public void start(PluginClassLoaderRegistry registry) {
        Map<String, Object> configMap = new HashMap<String, Object>();

        String systemPackages = "org.kantega.jexmec," +
                        "org.osgi.framework; version=1.3.0," +
                        "org.osgi.service.packageadmin; version=1.2.0," +
                        "org.osgi.service.startlevel; version=1.0.0," +
                        "org.osgi.service.url; version=1.0.0";
        for(String systemPackage : this.systemPackages) {
            systemPackages += "," + systemPackage;
        }
        configMap.put(Constants.FRAMEWORK_SYSTEMPACKAGES,
                        systemPackages);
        /*
        configMap.put(BundleCache.CACHE_PROFILE_DIR_PROP, "cache");
        */

        List<BundleActivator> list = new ArrayList<BundleActivator>();
        list.add(new AutoActivator(configMap));
        list.add(new PluginTrackerActivator(registry));
        configMap.put(FelixConstants.SYSTEMBUNDLE_ACTIVATORS_PROP, list);

        try {
            m_felix = new Felix(configMap);
            m_felix.start();
        }
        catch (Exception ex) {
            throw new RuntimeException(ex);
        }

    }

    public void stop() {
        try {
            m_felix.stop();
            m_felix.waitForStop(10000);
        } catch (BundleException e) {
            throw new RuntimeException(e);
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

    public Felix getFelix() {
        return m_felix;
    }

    class PluginTrackerActivator implements BundleActivator, SynchronousBundleListener {

        private final Map<Long, ClassLoader> loaders = new HashMap<Long, ClassLoader>();
        private final PluginClassLoaderRegistry registry;

        public PluginTrackerActivator(PluginClassLoaderRegistry registry) {
            this.registry = registry;
        }

        public void start(BundleContext bundleContext) throws Exception {
            bundleContext.addBundleListener(this);

        }

        public void stop(BundleContext bundleContext) throws Exception {
            bundleContext.removeBundleListener(this);
        }

        public void bundleChanged(BundleEvent bundleEvent) {
            final int eventType = bundleEvent.getType();
            final long bundleId = bundleEvent.getBundle().getBundleId();

            if (BundleEvent.STARTED == eventType) {
                ClassLoader bundleDelegatingClassLoader = createBundleClassLoader(bundleEvent.getBundle());

                registry.addClassLoader(bundleDelegatingClassLoader);
                loaders.put(bundleId, bundleDelegatingClassLoader);

            } else if (BundleEvent.STOPPING == eventType) {
                ClassLoader loader = loaders.get(bundleId);
                registry.removeClassLoader(loader);
                loaders.remove(bundleId);
            }

        }
    }

    private ClassLoader createBundleClassLoader(final Bundle bundle) {
        return new BundleDelegatingClassLoader(bundle);
    }

    private static class BundleDelegatingClassLoader extends ClassLoader {
        private final Bundle bundle;

        BundleDelegatingClassLoader(Bundle bundle) {
            this.bundle = bundle;
        }

        @Override
        protected Class<?> findClass(String name) throws ClassNotFoundException {
            return bundle.loadClass(name);
        }

        @Override
        protected URL findResource(String name) {
            return bundle.getResource(name);
        }

        @Override
        protected Enumeration<URL> findResources(String name) throws IOException {

            @SuppressWarnings("unchecked")
            Enumeration<URL> resources = bundle.getResources(name);
            return resources;
        }
    }
}
